#include "stdafx.h"
#include "ScanFinger.h"
#include "MainFrm.h"
#include "ScanFingerDoc.h"
#include "ScanFingerView.h"

//#ifdef _DEBUG
//#define new DEBUG_NEW
//#endif


#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CScanFingerApp, CWinApp) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
  ON_COMMAND(ID_APP_ABOUT, &CScanFingerApp::OnAppAbout)
  ON_COMMAND(ID_FILE_NEW, &CWinApp::OnFileNew)
  ON_COMMAND(ID_FILE_OPEN, &CWinApp::OnFileOpen)
END_MESSAGE_MAP()


#pragma warning(suppress: 26455)
CScanFingerApp::CScanFingerApp() : m_hScanSession{0},
                                   m_ScanState{State::None},
                                   m_ScanHR{S_OK},
                                   m_nGDIPlusToken{0}
{
  SetAppID(_T("ScanFinger.AppID.NoVersion"));
}

//The one and only CScanFingerApp object
#pragma warning(suppress: 26426)
CScanFingerApp theApp;

BOOL CScanFingerApp::InitInstance()
{
  INITCOMMONCONTROLSEX InitCtrls;
  InitCtrls.dwSize = sizeof(InitCtrls);
  InitCtrls.dwICC = ICC_WIN95_CLASSES;
  if (!InitCommonControlsEx(&InitCtrls))
    return FALSE;

  //Let the base class do its thing
  if (!__super::InitInstance())
    return FALSE;

  //Initialize GDI+
  const Gdiplus::GdiplusStartupInput gdiplusStartupInput{};
#pragma warning(suppress: 26486)
  if (Gdiplus::GdiplusStartup(&m_nGDIPlusToken, &gdiplusStartupInput, nullptr) != Gdiplus::Ok)
  {
    AfxMessageBox(_T("Failed to initialize GDI+"), MB_OK | MB_ICONSTOP);
    return FALSE;
  }

  EnableTaskbarInteraction(FALSE);
  SetRegistryKey(_T("Local AppWizard-Generated Applications"));

  //Register the application's document templates.  Document templates
  //serve as the connection between documents, frame windows and views
#pragma warning(suppress: 26400 26403 26409)
  auto pDocTemplate{new CSingleDocTemplate(IDR_MAINFRAME, RUNTIME_CLASS(CScanFingerDoc),
                                           RUNTIME_CLASS(CMainFrame), RUNTIME_CLASS(CScanFingerView))};
  if (pDocTemplate == nullptr)
    return FALSE;
  AddDocTemplate(pDocTemplate);

  //Parse command line for standard shell commands, DDE, file open
  CCommandLineInfo cmdInfo;
  ParseCommandLine(cmdInfo);

  //Dispatch commands specified on the command line.  Will return FALSE if
  //app was launched with /RegServer, /Register, /Unregserver or /Unregister.
  if (!ProcessShellCommand(cmdInfo))
    return FALSE;

  //The one and only window has been initialized, so show and update it
  m_pMainWnd->ShowWindow(SW_SHOW);
  m_pMainWnd->UpdateWindow();

  StartNewScan(true);

  return TRUE;
}

class CAboutDlg : public CDialog
{
public:
//Constructors / Destructors
  CAboutDlg();
  CAboutDlg(const CAboutDlg&) = delete;
  CAboutDlg(CAboutDlg&&) = delete;
  ~CAboutDlg() = default; //NOLINT(modernize-use-override)

//Methods
  CAboutDlg& operator=(const CAboutDlg&) = delete;
  CAboutDlg& operator=(CAboutDlg&&) = delete;

//Dialog Data
  enum { IDD = IDD_ABOUTBOX };

protected:
//Methods
  void DoDataExchange(CDataExchange* pDX) override;

  DECLARE_MESSAGE_MAP()
};

#pragma warning(suppress: 26455)
CAboutDlg::CAboutDlg() : CDialog{CAboutDlg::IDD}
{
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
  __super::DoDataExchange(pDX);
}

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CAboutDlg, CDialog) //NOLINT(clang-diagnostic-unused-local-typedef, modernize-avoid-c-arrays)
END_MESSAGE_MAP()

void CScanFingerApp::OnAppAbout()
{
  CAboutDlg aboutDlg;
  aboutDlg.DoModal();
}

int CScanFingerApp::ExitInstance()
{
  if (m_hScanSession != 0)
  {
    //Cancel the scan if in progress
    if (m_ScanState == State::Scanning)
    {
      WinBioCancel(m_hScanSession);
      WinBioWait(m_hScanSession);
    }

    //Close up the session
    WinBioCloseSession(m_hScanSession);
    m_hScanSession = 0;
  }

  //Shutdown GDI+
  Gdiplus::GdiplusShutdown(m_nGDIPlusToken);

  //Let the base class do its thing
  return __super::ExitInstance();
}

VOID CALLBACK CScanFingerApp::_CaptureSampleCallback(_In_opt_ PVOID CaptureCallbackContext, _In_ HRESULT OperationStatus, _In_ WINBIO_UNIT_ID UnitId, _In_reads_bytes_(nSampleSize) PWINBIO_BIR pSample, _In_ SIZE_T nSampleSize, _In_ WINBIO_REJECT_DETAIL RejectDetail)
{
#pragma warning(suppress: 26429)
  auto pThis{static_cast<CScanFingerApp*>(CaptureCallbackContext)};
  ASSERT(pThis != nullptr);

  pThis->CaptureSampleCallback(OperationStatus, UnitId, pSample, nSampleSize, RejectDetail);
}

void CScanFingerApp::CaptureSampleCallback(_In_ HRESULT OperationStatus, _In_ WINBIO_UNIT_ID UnitId, _In_reads_bytes_(nSampleSize) PWINBIO_BIR pSample, _In_ SIZE_T nSampleSize, _In_ WINBIO_REJECT_DETAIL RejectDetail)
{
  //Reset the state variables
  m_sScanInfoMessage.Empty();
  if (m_pScanImage != nullptr)
    m_pScanImage.reset();
  m_ScanHR = OperationStatus;
  m_sScanImageMessage.Empty();

  if (FAILED(m_ScanHR))
  {
    //Report the capture error
    if (m_ScanHR == WINBIO_E_BAD_CAPTURE)
      m_sScanInfoMessage.AppendFormat(_T("Failed to capture the fingerprint scan.\r\nReason:\"%s\""), RejectDetailToString(RejectDetail).GetString());
    else
    {
      _com_error error(OperationStatus);
      m_sScanInfoMessage.AppendFormat(_T("Failed to capture the fingerprint scan, Error:%08X, Description:\"%s\""), m_ScanHR, CString(error.Description().operator LPCWSTR()).GetString());
    }

    //Free the sample
    if (pSample != nullptr)
    {
      WinBioFree(pSample);
      pSample = nullptr;
    }
  }
  else
  {
    //Start the dump of the details
    m_sScanInfoMessage.Append(_T("** CAPTURED SCAN DETAILS **\r\n\r\n"));
    m_sScanInfoMessage.AppendFormat(_T("Acquired sample from unit: %u\r\n"), UnitId);
#pragma warning(suppress: 26472)
    m_sScanInfoMessage.AppendFormat(_T("Sample size (Bytes): %d\r\n"), static_cast<int>(nSampleSize));
#pragma warning(suppress: 26496)
    AFXASSUME(pSample != nullptr);
    m_sScanInfoMessage.AppendFormat(_T("VendorDataBlock size (Bytes): %u\r\n"), pSample->VendorDataBlock.Size);
    m_sScanInfoMessage.AppendFormat(_T("SignatureBlock size (Bytes): %u\r\n\r\n"), pSample->SignatureBlock.Size);

    //Dump details from the header block
    m_sScanInfoMessage.AppendFormat(_T("HeaderBlock size (Bytes): %u\r\n"), pSample->HeaderBlock.Size);
    if (pSample->HeaderBlock.Size)
    {
#pragma warning(suppress: 26481 26490)
      auto BirHeader{reinterpret_cast<const WINBIO_BIR_HEADER*>((reinterpret_cast<PBYTE>(pSample)) + pSample->HeaderBlock.Offset)};
      m_sScanInfoMessage.AppendFormat(_T(" ValidFields: 0x%08X\r\n"), static_cast<int>(BirHeader->ValidFields));
      m_sScanInfoMessage.AppendFormat(_T(" HeaderVersion: 0x%08X\r\n"), static_cast<int>(BirHeader->HeaderVersion));
      m_sScanInfoMessage.AppendFormat(_T(" PatronHeaderVersion: %d\r\n"), static_cast<int>(BirHeader->PatronHeaderVersion));
      m_sScanInfoMessage.AppendFormat(_T(" DataFlags: %d\r\n"), static_cast<int>(BirHeader->DataFlags));
      m_sScanInfoMessage.AppendFormat(_T(" Type: %u\r\n"), BirHeader->Type);
      m_sScanInfoMessage.AppendFormat(_T(" Type: %d\r\n"), static_cast<int>(BirHeader->Subtype));
      m_sScanInfoMessage.AppendFormat(_T(" Purpose: %d\r\n"), static_cast<int>(BirHeader->Purpose));
      m_sScanInfoMessage.AppendFormat(_T(" DataQuality: %d\r\n"), static_cast<int>(BirHeader->DataQuality));
      m_sScanInfoMessage.AppendFormat(_T(" CreationDate: %I64d\r\n"), BirHeader->CreationDate.QuadPart);
      m_sScanInfoMessage.AppendFormat(_T(" ValidityPeriod BeginDate: %I64d\r\n"), BirHeader->ValidityPeriod.BeginDate.QuadPart);
      m_sScanInfoMessage.AppendFormat(_T(" ValidityPeriod EndDate: %I64d\r\n"), BirHeader->ValidityPeriod.EndDate.QuadPart);
      m_sScanInfoMessage.AppendFormat(_T(" BiometricDataFormat Owner: %d\r\n"), static_cast<int>(BirHeader->BiometricDataFormat.Owner));
      m_sScanInfoMessage.AppendFormat(_T(" BiometricDataFormat Type: %d\r\n"), static_cast<int>(BirHeader->BiometricDataFormat.Type));
      m_sScanInfoMessage.AppendFormat(_T(" ProductId Owner: %d\r\n"), static_cast<int>(BirHeader->ProductId.Owner));
      m_sScanInfoMessage.AppendFormat(_T(" ProductId Type: %d\r\n"), static_cast<int>(BirHeader->ProductId.Type));
    }

    //Dump details from the standard data block
    m_sScanInfoMessage.AppendFormat(_T("\r\nStandardDataBlock size (Bytes): %u\r\n"), pSample->StandardDataBlock.Size);
    if (pSample->StandardDataBlock.Size)
    {
#pragma warning(suppress: 26481 26490)
      auto AnsiBdbHeader{reinterpret_cast<const WINBIO_BDB_ANSI_381_HEADER*>((reinterpret_cast<PBYTE>(pSample)) + pSample->StandardDataBlock.Offset)};
      m_sScanInfoMessage.AppendFormat(_T(" Record length: %I64u\r\n"), AnsiBdbHeader->RecordLength);
      m_sScanInfoMessage.AppendFormat(_T(" FormatIdentifier: 0x%08X\r\n"), AnsiBdbHeader->FormatIdentifier);
      m_sScanInfoMessage.AppendFormat(_T(" VersionNumber: 0x%08X\r\n"), AnsiBdbHeader->VersionNumber);
      m_sScanInfoMessage.AppendFormat(_T(" ProductId Owner: %d\r\n"), static_cast<int>(AnsiBdbHeader->ProductId.Owner));
      m_sScanInfoMessage.AppendFormat(_T(" ProductId Type: %d\r\n"), static_cast<int>(AnsiBdbHeader->ProductId.Type));
      m_sScanInfoMessage.AppendFormat(_T(" CaptureDeviceId: %d\r\n"), static_cast<int>(AnsiBdbHeader->CaptureDeviceId));
      m_sScanInfoMessage.AppendFormat(_T(" ImageAcquisitionLevel: %d\r\n"), static_cast<int>(AnsiBdbHeader->ImageAcquisitionLevel));
      m_sScanInfoMessage.AppendFormat(_T(" HorizontalScanResolution: %d\r\n"), static_cast<int>(AnsiBdbHeader->HorizontalScanResolution));
      m_sScanInfoMessage.AppendFormat(_T(" VerticalScanResolution: %d\r\n"), static_cast<int>(AnsiBdbHeader->VerticalScanResolution));
      m_sScanInfoMessage.AppendFormat(_T(" HorizontalImageResolution: %d\r\n"), static_cast<int>(AnsiBdbHeader->HorizontalImageResolution));
      m_sScanInfoMessage.AppendFormat(_T(" VerticalImageResolution: %d\r\n"), static_cast<int>(AnsiBdbHeader->VerticalImageResolution));
      m_sScanInfoMessage.AppendFormat(_T(" ElementCount: %d\r\n"), static_cast<int>(AnsiBdbHeader->ElementCount));
      m_sScanInfoMessage.AppendFormat(_T(" ScaleUnits: %d\r\n"), static_cast<int>(AnsiBdbHeader->ScaleUnits));
      m_sScanInfoMessage.AppendFormat(_T(" PixelDepth: %d\r\n"), static_cast<int>(AnsiBdbHeader->PixelDepth));
      m_sScanInfoMessage.AppendFormat(_T(" ImageCompressionAlg: %d\r\n"), static_cast<int>(AnsiBdbHeader->ImageCompressionAlg));

      if (AnsiBdbHeader->ElementCount)
      {
#pragma warning(suppress: 26481 26490)
        auto AnsiBdbRecord{reinterpret_cast<const WINBIO_BDB_ANSI_381_RECORD*>((reinterpret_cast<const BYTE*>(AnsiBdbHeader)) + sizeof(WINBIO_BDB_ANSI_381_HEADER))};
        m_sScanInfoMessage.Append(_T(" ANSI 281 RECORD\r\n"));
        m_sScanInfoMessage.AppendFormat(_T("  BlockLength: %u\r\n"), AnsiBdbRecord->BlockLength);
        m_sScanInfoMessage.AppendFormat(_T("  HorizontalLineLength: %u\r\n"), AnsiBdbRecord->HorizontalLineLength);
        m_sScanInfoMessage.AppendFormat(_T("  VerticalLineLength: %u\r\n"), AnsiBdbRecord->VerticalLineLength);
        m_sScanInfoMessage.AppendFormat(_T("  Position: %d\r\n"), static_cast<int>(AnsiBdbRecord->Position));
        m_sScanInfoMessage.AppendFormat(_T("  CountOfViews: %d\r\n"), static_cast<int>(AnsiBdbRecord->CountOfViews));
        m_sScanInfoMessage.AppendFormat(_T("  ViewNumber: %d\r\n"), static_cast<int>(AnsiBdbRecord->ViewNumber));
        m_sScanInfoMessage.AppendFormat(_T("  ImageQuality: %d\r\n"), static_cast<int>(AnsiBdbRecord->ImageQuality));
        m_sScanInfoMessage.AppendFormat(_T("  ImpressionType: %d\r\n"), static_cast<int>(AnsiBdbRecord->ImpressionType));
        const size_t nImageSize{AnsiBdbRecord->BlockLength - sizeof(WINBIO_BDB_ANSI_381_RECORD)};
        if (nImageSize > 0)
        {
          if (AnsiBdbHeader->PixelDepth == 8)
          {
            //Create the image from the scan
            const int nWidth{AnsiBdbRecord->HorizontalLineLength};
            const int nHeight{AnsiBdbRecord->VerticalLineLength};
            m_pScanImage = std::make_unique<Gdiplus::Bitmap>(nWidth, nHeight);
            if (m_pScanImage->GetLastStatus() != Gdiplus::Ok)
            {
              //Report the failure
              m_sScanInfoMessage.AppendFormat(_T("\r\nFailed to create scan image, Error:%d"), m_pScanImage->GetLastStatus());
              m_pScanImage.reset();
            }
            else
            {
#pragma warning(suppress: 26481 26429 26490)
              const BYTE* pbyFirstPixel{reinterpret_cast<const BYTE*>(AnsiBdbRecord) + sizeof(WINBIO_BDB_ANSI_381_RECORD)};
              for (int y{0}; y<nHeight; ++y)
              {
                for (int x{0}; x<nWidth; ++x)
                {
#pragma warning(suppress: 26481)
                  const BYTE byPixel{pbyFirstPixel[(y * nWidth) + x]};
                  m_pScanImage->SetPixel(x, y, Gdiplus::Color(byPixel, byPixel, byPixel));
                }
              }
            }
          }
          else
            m_sScanInfoMessage.Append(_T("\r\nAn 8 bit image was not provided\r\n"));
        }
        else
          m_sScanInfoMessage.Append(_T("\r\nNo image was provided\r\n"));
      }
    }

    //Free the sample
    if (pSample != nullptr)
    {
      WinBioFree(pSample);
      pSample = nullptr;
    }
  }

  //Update our state
  m_ScanState = State::ScanAcquired;

  //Update the UI
  CWnd* pWnd{m_pMainWnd};
  if ((pWnd != nullptr) && IsWindow(pWnd->GetSafeHwnd()))
    pWnd->PostMessage(WM_ACQUIRED_SAMPLE, m_ScanHR);
}

CString CScanFingerApp::GetErrorString(_In_ HRESULT hr)
{
  //Lookup the error using FormatMessage
  CString sError;
  LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26490)
  const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_HMODULE,
                                     nullptr, HRESULT_CODE(hr), MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
                                     reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};
  if (dwReturn != 0)
  {
    sError = lpBuffer;
    sError.TrimRight(_T("\r\n"));
    LocalFree(lpBuffer);
  }
  return sError;
}

CString CScanFingerApp::GetErrorStringWinBIO(_In_ HRESULT hr)
{
  //Lookup the error using FormatMessage
  CString sError;
  LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26462)
  HMODULE hWinBIO{GetModuleHandle(_T("WINBIO.DLL"))};
#pragma warning(suppress: 26490)
  const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_HMODULE,
                                     hWinBIO, hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
                                     reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};
  if (dwReturn != 0)
  {
    sError = lpBuffer;
    sError.TrimRight(_T("\r\n"));
    LocalFree(lpBuffer);
  }
  else
    sError = GetErrorString(hr);
  return sError;
}

CString CScanFingerApp::RejectDetailToString(_In_ WINBIO_REJECT_DETAIL rejectDetail)
{
  CString sError;
  switch (rejectDetail)
  {
    case WINBIO_FP_TOO_HIGH:
    {
      sError = _T("Too high");
      break;
    }
    case WINBIO_FP_TOO_LOW:
    {
      sError = _T("Too low");
      break;
    }
    case WINBIO_FP_TOO_LEFT:
    {
      sError = _T("Too left");
      break;
    }
    case WINBIO_FP_TOO_RIGHT:
    {
      sError = _T("Too right");
      break;
    }
    case WINBIO_FP_TOO_FAST:
    {
      sError = _T("Too fast");
      break;
    }
    case WINBIO_FP_TOO_SLOW:
    {
      sError = _T("Too slow");
      break;
    }
    case WINBIO_FP_POOR_QUALITY:
    {
      sError = _T("Poor quality");
      break;
    }
    case WINBIO_FP_TOO_SKEWED:
    {
      sError = _T("Too skewed");
      break;
    }
    case WINBIO_FP_TOO_SHORT:
    {
      sError = _T("Too short");
      break;
    }
    case WINBIO_FP_MERGE_FAILURE:
    {
      sError = _T("Merge failure");
      break;
    }
    default:
    {
      break;
    }
  }

  return sError;
}

bool CScanFingerApp::StartNewScan(_In_ bool bRepromptUserToScan)
{
  //Cancel the scan if in progress
  if (m_ScanState == State::Scanning)
  {
    WinBioCancel(m_hScanSession);
    WinBioWait(m_hScanSession);
  }

#pragma warning(suppress: 26466)
  auto pFrameWnd{static_cast<CFrameWnd*>(m_pMainWnd)};
  if (pFrameWnd == nullptr)
    return FALSE;
  ASSERT(pFrameWnd->IsKindOf(RUNTIME_CLASS(CFrameWnd)));
#pragma warning(suppress: 26429 26490)
  auto pDoc{reinterpret_cast<CScanFingerDoc*>(pFrameWnd->GetActiveDocument())};
  ASSERT(pDoc != nullptr);

  //Create a new WinBIO session if necessary
  if (m_hScanSession == 0)
  {
    m_ScanHR = WinBioOpenSession(WINBIO_TYPE_FINGERPRINT, WINBIO_POOL_SYSTEM, WINBIO_FLAG_RAW, nullptr, 0, nullptr, &m_hScanSession);
    if (FAILED(m_ScanHR))
    {
      if (m_ScanHR == E_ACCESSDENIED)
        m_sScanInfoMessage.Format(_T("An access denied error occurred when trying to open the fingerprint scanner.\r\nYou should rerun this application elevated to work correctly"));
      else
        m_sScanInfoMessage.Format(_T("Failed to open fingerprint scanner.\r\nError:0x%08X, Description:\"%s\""), m_ScanHR, GetErrorStringWinBIO(m_ScanHR).GetString());
      m_ScanState = State::ErrorOpening;
      pDoc->UpdateAllViews(nullptr);
      return TRUE;
    }
  }

  //Acquire the sample
  m_ScanHR = WinBioCaptureSampleWithCallback(m_hScanSession, WINBIO_NO_PURPOSE_AVAILABLE, WINBIO_DATA_FLAG_RAW, _CaptureSampleCallback, this);
  if (FAILED(m_ScanHR))
  {
    m_sScanInfoMessage.Format(_T("Failed to acquire the fingerprint scan.\r\nError:0x%08X, Description:\"%s\""), m_ScanHR, GetErrorStringWinBIO(m_ScanHR).GetString());
    m_ScanState = State::ErrorOpening;
    WinBioCloseSession(m_hScanSession);
    m_hScanSession = 0;
    pDoc->UpdateAllViews(nullptr);
    return TRUE;
  }

  //Reset the image and text if necessary
  if (bRepromptUserToScan)
  {
    if (m_pScanImage != nullptr)
      m_pScanImage.reset();
    m_sScanInfoMessage = _T("Please scan your finger on the fingerprint scanner now.\r\n\r\nWhen the scan is captured this window will contain detailed information about the fingerprint scan");
    m_sScanImageMessage = _T("Please scan your finger on the fingerprint scanner now\r\n\r\nWhen the scan is captured this window will display the scan as an image");
  }

  //Update our state
  m_ScanState = State::Scanning;

  //Update all the views
  pDoc->UpdateAllViews(nullptr);

  return TRUE;
}
