#pragma once


#include "resource.h" //main symbols

static constexpr UINT WM_ACQUIRED_SAMPLE = WM_USER + 1;


class CScanFingerApp : public CWinApp
{
public:
//Constructors / Destructors
  CScanFingerApp();
  CScanFingerApp(const CScanFingerApp&) = delete;
  CScanFingerApp(CScanFingerApp&&) = delete;
  ~CScanFingerApp() = default; //NOLINT(modernize-use-override)

//Enums
  enum class State
  {
    None,
    ErrorOpening,
    Scanning,
    ScanAcquired
  };

//Methods
  CScanFingerApp& operator=(const CScanFingerApp&) = delete;
  CScanFingerApp& operator=(CScanFingerApp&&) = delete;
  BOOL InitInstance() override;
  int ExitInstance() override;
  bool StartNewScan(_In_ bool bRepromptUserToScan);

//Member variables
  WINBIO_SESSION_HANDLE m_hScanSession;
  std::atomic<State> m_ScanState;
  std::unique_ptr<Gdiplus::Bitmap> m_pScanImage;
  HRESULT m_ScanHR;
  CString m_sScanInfoMessage;
  CString m_sScanImageMessage;
  ULONG_PTR m_nGDIPlusToken;

protected:
//Methods
  static CString GetErrorString(_In_ HRESULT hr);
  static CString GetErrorStringWinBIO(_In_ HRESULT hr);
  static CString RejectDetailToString(_In_ WINBIO_REJECT_DETAIL rejectDetail);
  static VOID CALLBACK _CaptureSampleCallback(_In_opt_ PVOID CaptureCallbackContext, _In_ HRESULT OperationStatus, _In_ WINBIO_UNIT_ID UnitId, _In_reads_bytes_(nSampleSize) PWINBIO_BIR pSample, _In_ SIZE_T nSampleSize, _In_ WINBIO_REJECT_DETAIL RejectDetail);
  void CaptureSampleCallback(_In_ HRESULT OperationStatus, _In_ WINBIO_UNIT_ID UnitId, _In_reads_bytes_(nSampleSize) PWINBIO_BIR pSample, _In_ SIZE_T nSampleSize, _In_ WINBIO_REJECT_DETAIL RejectDetail);

//Message handlers
  afx_msg void OnAppAbout();

  DECLARE_MESSAGE_MAP()
};

extern CScanFingerApp theApp;
